from json import loads as json_loads

import pytest

from gpt_json.transformations import fix_truncated_json, fix_bools


@pytest.mark.parametrize(
    "input_string,expected",
    [
        ('{ "key": "value", ', True),
        ('[{"key": "value"}, ', True),
        ('{ "key": "value"}', False),
        ('[{"key": "value"}]', False),
        ('{"key": "value",', True),
        ('random_string', False)
    ]
)
def test_is_truncated(input_string: str, expected: bool):
    _, is_truncated = fix_truncated_json(input_string)
    assert is_truncated == expected


@pytest.mark.parametrize(
    "broken_string,expected",
    [
        ('{ "key1": "value1", "key2": "value2", ', { "key1": "value1", "key2": "value2"}),
        ('[{"key1": "value1"}, {"key2": "value2"}, ', [{"key1": "value1"}, {"key2": "value2"}]),
        ('{ "key1": "value1", "key2": { "nestedKey": "nestedValue",', { "key1": "value1", "key2": { "nestedKey": "nestedValue"}}),
        ('[{"key1": "value1"}, {"key2": "value2"}, {"key3": "value3", "key4": "value4",', [{"key1": "value1"}, {"key2": "value2"}, {"key3": "value3", "key4": "value4"}]),
        ('[{"key1": "value1"}, {"key2": [1, 2, 3,', [{"key1": "value1"}, {"key2": [1, 2, 3]}]),
        ('{ "key1": "value1", "key2": "value2"}', { "key1": "value1", "key2": "value2"}),
        ('[{"key1": "value1"}, {"key2": "value2"}]', [{"key1": "value1"}, {"key2": "value2"}]),
        # test case with an escaped quote
        ('[{"key1": "value1"}, {"key2": "value2 \\" and a quote', [{"key1": "value1"}, {"key2": "value2 \" and a quote"}]),
        ('[{"key1": "value1"}, {"key2": "value2 start\n\n', [{"key1": "value1"}, {"key2": "value2 start"}]),
        # observed examples
        ('[{"key1": [123]', [{"key1": [123]}]),
        ('{"key1": [\n"abc",\n "def', {"key1": ["abc", "def"]}),
    ]
)
def test_fix_truncated_json(broken_string, expected):
    fixed_string, _ = fix_truncated_json(broken_string)

    print("BROKEN", broken_string)
    print("EXPECTED", expected)
    print("ACTUAL", fixed_string)

    assert json_loads(fixed_string) == expected

@pytest.mark.parametrize(
    "input_str, expected_output",
    [
        # No change required
        (
        '{"status": "success", "data": {"isAdmin": true, "isActive": false}}',
        ('{"status": "success", "data": {"isAdmin": true, "isActive": false}}', False)
        ),
        # Fix "True" to "true"
        (
        '{"status": "success", "data": {"isAdmin": True, "isActive": false}}',
        ('{"status": "success", "data": {"isAdmin": true, "isActive": false}}', True)
        ),
        # Fix "False" to "false"
        (
        '{"status": "success", "data": {"isAdmin": true, "isActive": False}}',
        ('{"status": "success", "data": {"isAdmin": true, "isActive": false}}', True)
        ),
        # Fix both "True" and "False"
        (
        '{"status": "success", "data": {"isAdmin": True, "isActive": False}}',
        ('{"status": "success", "data": {"isAdmin": true, "isActive": false}}', True)
        ),
        # Test with string containing "True" and "False" as substrings
        (
        '{"status": "TrueValue", "data": {"isAdmin": True, "isActive": False, "attribute": "FalsePositive"}}',
        ('{"status": "TrueValue", "data": {"isAdmin": true, "isActive": false, "attribute": "FalsePositive"}}', True)
        ),
    ]
)
def test_fix_bools(input_str, expected_output):
    assert fix_bools(input_str) == expected_output



import logging
from dataclasses import replace
from json import loads as json_loads
from json.decoder import JSONDecodeError
from typing import Any, Generic, List, Type, TypeVar, get_args, get_origin

import backoff
import openai
from openai.error import APIConnectionError, RateLimitError
from openai.error import Timeout as OpenAITimeout
from pydantic import BaseModel
from tiktoken import encoding_for_model

from gpt_json.models import (FixTransforms, GPTMessage, GPTModelVersion,
                             ResponseType)
from gpt_json.parsers import find_json_response
from gpt_json.prompts import generate_schema_prompt
from gpt_json.transformations import fix_bools, fix_truncated_json

logger = logging.getLogger('gptjson_logger')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


def handle_backoff(details):
    logger.warning(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )

SchemaType = TypeVar("SchemaType", bound=BaseModel)

SCHEMA_PROMPT_TEMPLATE_KEY = "json_schema"

class GPTJSON(Generic[SchemaType]):
    """
    A wrapper over GPT that provides basic JSON parsing and response handling.

    """
    _cls_schema_model: Type[SchemaType] = None
    schema_model: Type[SchemaType] = None

    def __init__(
        self,
        api_key: str | None = None,
        model: GPTModelVersion | str = GPTModelVersion.GPT_4,
        auto_trim: bool = False,
        auto_trim_response_overhead: int = 0,
        # For messages that are relatively deterministic
        temperature = 0.2,
        timeout = 60,
        openai_max_retries=3,
        **kwargs,
    ):
        """
        :param api_key: OpenAI API key, if `OPENAI_API_KEY` environment variable is not set
        :param model: GPTModelVersion or string model name
        :param auto_trim: If True, automatically trim messages to fit within the model's token limit
        :param auto_trim_response_overhead: If auto_trim is True, will leave at least `auto_trim_response_overhead` space
            for the output payload. For GPT, initial prompt + response <= allowed tokens.
        :param temperature: Temperature (or variation) of response payload; 0 is the most deterministic, 1 is the most random
        :param timeout: Timeout in seconds for each OpenAI API calls
        :param openai_max_retries: Amount of times to retry failed API calls, caused by often transient load or rate limit issues
        :param kwargs: Additional arguments to pass to OpenAI's `openai.Completion.create` method

        """
        self.model = model.value if isinstance(model, GPTModelVersion) else model
        self.auto_trim = auto_trim
        self.temperature = temperature
        self.timeout = timeout
        self.openai_max_retries = openai_max_retries
        self.openai_arguments = kwargs
        self.schema_model = self._cls_schema_model
        GPTJSON._cls_schema_model = None

        if not self.schema_model:
            raise ValueError("GPTJSON needs to be instantiated with a schema model, like GPTJSON[MySchema](...args).")

        if get_origin(self.schema_model) in {list, List}:
            self.extract_type = ResponseType.LIST
        elif issubclass(self.schema_model, BaseModel):
            self.extract_type = ResponseType.DICTIONARY
        else:
            raise ValueError("GPTJSON needs to be instantiated with either a pydantic.BaseModel schema or a list of those schemas.")

        if self.auto_trim:
            if "gpt-4" in self.model:
                self.max_tokens = 8192 - auto_trim_response_overhead
            elif "gpt-3.5" in self.model:
                self.max_tokens = 4096 - auto_trim_response_overhead
            else:
                raise ValueError("Unknown model to infer max tokens, see https://platform.openai.com/docs/models/gpt-4 for more information on token length.")

        self.schema_prompt = generate_schema_prompt(self.schema_model)
        self.api_key = api_key

    async def run(
        self,
        messages: list[GPTMessage],
        max_response_tokens: int | None = None,
        format_variables: dict[str, Any] | None = None,
    ) -> tuple[SchemaType, FixTransforms] | tuple[None, None]:
        """
        :param messages: List of GPTMessage objects to send to the API
        :param max_response_tokens: Maximum number of tokens allowed in the response
        :param format_variables: Variables to format into the message template. Uses standard
            Python string formatting, like "Hello {name}".format(name="World")

        :return: Tuple of (parsed response, fix transforms). The transformations here is a object that
            contains the allowed modifications that we might do to cleanup a GPT payload. It allows client callers
            to decide whether they want to allow the modifications or not.

        """
        messages = [
            self.fill_message_template(message, format_variables or {})
            for message in messages
        ]

        # Most requests succeed on the first try but we wrap it locally here in case
        # there is some temporarily instability with the API. If there are longer periods
        # of instability, there should be system-wide retries in a daemon.
        backoff_request_submission = backoff.on_exception(
            backoff.expo,
            (RateLimitError, OpenAITimeout, APIConnectionError),
            max_tries=self.openai_max_retries,
            on_backoff=handle_backoff
        )(self.submit_request)

        response = await backoff_request_submission(messages, max_response_tokens=max_response_tokens)
        logger.debug("------- RAW RESPONSE ----------")
        logger.debug(response["choices"])
        logger.debug("------- END RAW RESPONSE ----------")
        extracted_json, fixed_payload = self.extract_json(response, self.extract_type)

        # Cast to schema model
        if extracted_json is None:
            return None, None

        # Allow pydantic to handle the validation
        if isinstance(extracted_json, list):
            model = get_args(self.schema_model)[0]
            return [model(**item) for item in extracted_json], fixed_payload
        else:
            return self.schema_model(**extracted_json), fixed_payload

    def extract_json(self, completion_response, extract_type: ResponseType):
        """
        Assumes one main block of results, either list of dictionary

        """
        choices = completion_response["choices"]

        if not choices:
            logger.warning("No choices available, should report error...")
            return None, None

        full_response = choices[0]["message"]["content"]

        extracted_response = find_json_response(full_response, extract_type)
        if extracted_response is None:
            return None, None

        # Save the original response before we start modifying it
        fixed_response = extracted_response
        fixed_response, fixed_truncation = fix_truncated_json(fixed_response)
        fixed_response, fixed_bools = fix_bools(fixed_response)

        fixed_payload = FixTransforms(
            fixed_bools=fixed_bools,
            fixed_truncation=fixed_truncation,
        )

        try:
            return json_loads(fixed_response), fixed_payload
        except JSONDecodeError as e:
            logger.debug(f"Extracted: {extracted_response}")
            logger.debug(f"Did parse: {fixed_response}")
            logger.error(f"JSON decode error, likely malformed json input: {e}")
            return None, fixed_payload

    async def submit_request(
        self,
        messages: list[GPTMessage],
        max_response_tokens: int | None,
    ):
        logger.debug("------- START MESSAGE ----------")
        logger.debug(messages)
        logger.debug("------- END MESSAGE ----------")
        if self.auto_trim:
            messages = self.trim_messages(messages, self.max_tokens)

        optional_parameters = {}

        if max_response_tokens:
            optional_parameters["max_tokens"] = max_response_tokens

        return await openai.ChatCompletion.acreate(
            model=self.model,
            messages=[
                self.message_to_dict(message)
                for message in messages
            ],
            temperature=self.temperature,
            timeout=self.timeout,
            api_key=self.api_key,
            **optional_parameters,
            **self.openai_arguments,
        )

    def fill_message_template(self, message: GPTMessage, format_variables: dict[str, Any]):
        auto_format = {
            SCHEMA_PROMPT_TEMPLATE_KEY: self.schema_prompt,
        }

        # Regular quotes should passthrough to the next stage, except for our special keys
        content = message.content.replace('{', '{{').replace('}', '}}')
        for key in auto_format.keys():
            content = content.replace("{{" + key + "}}", "{" + key + "}")
        content = content.format(**auto_format)

        # We do this formatting in a separate pass so we can fill any template variables that might
        # have been left in the pydantic field typehints
        content = content.format(**format_variables)

        return replace(
            message,
            content=content,
        )

    def message_to_dict(self, message: GPTMessage):
        return {
            "role": message.role.value,
            "content": message.content
        }

    def trim_messages(self, messages: list[GPTMessage], n: int):
        """
        Returns a list of messages with a total token count less than n tokens,
        cropping the last message if needed.

        Args:
            messages (list): List of strings to be checked.
            n (int): The maximum number of tokens allowed.

        Returns:
            list: A list of messages with a total token count less than n tokens.
        """
        message_text = [
            message.content
            for message in messages
        ]

        enc = encoding_for_model("gpt-4")
        filtered_messages = []
        current_token_count = 0
        original_token_count = sum([
            len(enc.encode(message))
            for message in message_text
        ])

        for message in message_text:
            tokens = enc.encode(message)
            message_token_count = len(tokens)

            if current_token_count + message_token_count < n:
                filtered_messages.append(message)
                current_token_count += message_token_count
            else:
                remaining_tokens = n - current_token_count
                if remaining_tokens > 0:
                    cropped_message = enc.decode(tokens[:remaining_tokens])
                    filtered_messages.append(cropped_message)
                current_token_count += remaining_tokens
                break

        # Recreate the messages with our new text
        new_messages = [
            replace(
                messages[i],
                content=content,
            )
            for i, content in enumerate(filtered_messages)
        ]

        # Log a copy of the message array if we have to crop it
        if current_token_count != original_token_count:
            logger.debug(
                f"Trimmed message from {original_token_count} to {current_token_count} tokens",
                new_messages,
            )
        else:
            logger.debug(f"Skipping trim ({original_token_count}) ({current_token_count})")

        return new_messages

    def __class_getitem__(cls, item):
        new_cls = super().__class_getitem__(item)
        new_cls._cls_schema_model = item
        return new_cls



from pydantic import BaseModel, Field


class MySubSchema(BaseModel):
    name: str


class MySchema(BaseModel):
    text: str
    items: list[str]
    numerical: int | float
    sub_element: MySubSchema
    reason: bool = Field(description="Explanation")



from types import UnionType
from typing import List, get_args, get_origin

from pydantic import BaseModel


def generate_schema_prompt(schema: BaseModel | list[BaseModel]) -> str:
    """
    Converts the pydantic schema into a text representation that can be embedded
    into the prompt payload.

    """
    def generate_payload(model):
        payload = []
        for key, value in model.__fields__.items():
            field_annotation = model.__annotations__[key]
            annotation_origin = get_origin(field_annotation)
            annotation_arguments = get_args(field_annotation)

            if annotation_origin in {list, List}:
                payload.append(f'"{key}": {annotation_arguments[0].__name__}[]')
            elif annotation_origin == UnionType:
                payload.append(f'"{key}": {" | ".join([arg.__name__.lower() for arg in annotation_arguments])}')
            elif issubclass(value.type_, BaseModel):
                payload.append(f'"{key}": {generate_payload(value.type_)}')
            else:
                payload.append(f'"{key}": {value.type_.__name__.lower()}')
            if value.field_info.description:
                payload[-1] += f' // {value.field_info.description}'
        # All brackets are double defined so they will passthrough a call to `.format()` where we
        # pass custom variables
        return '{{\n' + ',\n'.join(payload) + '\n}}'

    origin = get_origin(schema)
    args = get_args(schema)

    if origin == list:
        if len(args) > 1:
            raise ValueError("Only one list schema is supported at this time")

        return '[' + '\n'.join(
            [
                generate_payload(sub_schema) + ", // Repeat for as many objects as are relevant"
                for sub_schema in args
            ]
        ) + ']'
    else:
        return generate_payload(schema)



from re import DOTALL, finditer

from gpt_json.models import ResponseType
from gpt_json.transformations import is_truncated


def find_json_response(full_response, extract_type):
    """
    Takes a full response that might contain other strings and attempts to extract the JSON payload.
    Has support for truncated JSON where the JSON begins but the token window ends before the json is
    is properly closed.

    """
    # Deal with fully included responses as well as truncated responses that only have one
    if extract_type == ResponseType.LIST:
        extracted_responses = list(finditer(r"(\[[^\]]*$|\[.*\])", full_response, flags=DOTALL))
    elif extract_type == ResponseType.DICTIONARY:
        extracted_responses = list(finditer(r"({[^}]*$|{.*})", full_response, flags=DOTALL))
    else:
        raise ValueError("Unknown extract_type")

    if not extracted_responses:
        print(f"Unable to find any responses of the matching type `{extract_type}`: `{full_response}`")
        return None

    if len(extracted_responses) > 1:
        print("Unexpected response > 1, continuing anyway...", extracted_responses)

    extracted_response = extracted_responses[0]

    if is_truncated(extracted_response.group(0)):
        # Start at the same location and just expand to the end of the message
        extracted_response = full_response[extracted_response.start():]
    else:
        extracted_response = extracted_response.group(0)

    return extracted_response



import pytest
import gpt_json.models as models_file   
from json import loads as json_loads, dumps as json_dumps
from enum import Enum

@pytest.mark.parametrize("model_file", [models_file])
def test_serializable_enums(model_file):
    """
    All our enums should be serializable as JSON

    """
    found_enums = 0
    for obj in model_file.__dict__.values():
        if isinstance(obj, type) and issubclass(obj, Enum):
            found_enums += 1
            for enum_value in obj:
                assert enum_value.value == json_loads(json_dumps(enum_value))

    # Every file listed in pytest should have at least one enum
    assert found_enums > 0



from unittest.mock import MagicMock, patch

import openai
import pytest
from pydantic import BaseModel, Field

from gpt_json.gpt import GPTJSON
from gpt_json.models import (FixTransforms, GPTMessage, GPTMessageRole,
                             GPTModelVersion)
from gpt_json.tests.shared import MySchema, MySubSchema


def test_throws_error_if_no_model_specified():
    with pytest.raises(ValueError, match="needs to be instantiated with a schema model"):
        GPTJSON(None)


@pytest.mark.parametrize(
        "role_type,expected",
        [
            (GPTMessageRole.SYSTEM, "system"),
            (GPTMessageRole.USER, "user"),
            (GPTMessageRole.ASSISTANT, "assistant"),
        ]
)
def test_cast_message_to_gpt_format(role_type: GPTMessageRole, expected: str):
    parser = GPTJSON[MySchema](None)
    assert parser.message_to_dict(
        GPTMessage(
            role=role_type,
            content="test",
        )
    )["role"] == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "schema_typehint,response_raw,parsed,expected_transformations",
    [
        (
            MySchema,
            """
            Your response is as follows:
            {
                "text": "Test",
                "items": ["Item 1", "Item 2"],
                "numerical": 123,
                "sub_element": {
                    "name": "Test"
                },
                "reason": true
            }
            Your response is above.
            """,
            MySchema(
                text="Test",
                items=["Item 1", "Item 2"],
                numerical=123,
                sub_element=MySubSchema(
                    name="Test"
                ),
                reason=True,
            ),
            FixTransforms()
        ),
        (
            list[MySchema],
            """
            Your response is as follows:
            [
                {
                    "text": "Test",
                    "items": ["Item 1", "Item 2"],
                    "numerical": 123,
                    "sub_element": {
                        "name": "Test"
                    },
                    "reason": true
                }
            ]
            Your response is above.
            """,
            [
                MySchema(
                    text="Test",
                    items=["Item 1", "Item 2"],
                    numerical=123,
                    sub_element=MySubSchema(
                        name="Test"
                    ),
                    reason=True,
                )                
            ],
            FixTransforms()
        ),
        (
            MySchema,
            """
            Your response is as follows:
            {
                "text": "Test",
                "numerical": 123,
                "reason": True,
                "sub_element": {
                    "name": "Test"
                },
                "items": ["Item 1", "Item 2

            """,
            MySchema(
                text="Test",
                items=["Item 1", "Item 2"],
                numerical=123,
                sub_element=MySubSchema(
                    name="Test"
                ),
                reason=True,
            ),
            FixTransforms(fixed_bools=True, fixed_truncation=True),
        ),
    ]
)
async def test_acreate(schema_typehint, response_raw, parsed, expected_transformations: FixTransforms):
    model_version = GPTModelVersion.GPT_3_5
    messages = [
        GPTMessage(
            role=GPTMessageRole.USER,
            content="Input prompt",
        )
    ]

    model = GPTJSON[schema_typehint](
        None,
        model=model_version,
        temperature=0.0,
        timeout=60,
    )

    # Define mock response
    mock_response = {
        "choices": [
            {
                "message": {
                    "role": "assistant",
                    "content": response_raw,
                },
                "index": 0,
                "finish_reason": "stop",
            }
        ]
    }

    # Create the mock
    with patch.object(openai.ChatCompletion, "acreate") as mock_acreate:
        # Make the mock function asynchronous
        mock_acreate.return_value = mock_response

        # Call the function and pass the expected parameters
        response, transformations = await model.run(messages=messages)

        # Assert that the mock function was called with the expected parameters
        mock_acreate.assert_called_with(
            model=model_version.value,
            messages=[
                {
                    "role": message.role.value,
                    "content": message.content,
                }
                for message in messages
            ],
            temperature=0.0,
            timeout=60,
            api_key=None,
        )

    assert response == parsed
    assert transformations == expected_transformations


@pytest.mark.parametrize(
    "input_messages,expected_output_messages",
    [
        # Messages fit within max_tokens, no change expected
        (
            [
                GPTMessage(role=GPTMessageRole.SYSTEM, content="Hello"),
                GPTMessage(role=GPTMessageRole.USER, content="World!"),
            ],
            [
                GPTMessage(role=GPTMessageRole.SYSTEM, content="Hello"),
                GPTMessage(role=GPTMessageRole.USER, content="World!"),
            ],
        ),
        # All messages trimmed to fit max_tokens
        (
            [
                GPTMessage(role=GPTMessageRole.SYSTEM, content="Hello"),
                GPTMessage(role=GPTMessageRole.USER, content="World" * 10000),
            ],
            [
                GPTMessage(role=GPTMessageRole.SYSTEM, content="Hello"),
                GPTMessage(role=GPTMessageRole.USER, content="World" * (8192-1)),
            ],
        ),
    ],
)
def test_trim_messages(input_messages, expected_output_messages):
    gpt = GPTJSON[MySchema](None, auto_trim=True, auto_trim_response_overhead=0)

    output_messages = gpt.trim_messages(input_messages, n=8192)

    assert len(output_messages) == len(expected_output_messages)

    for output_message, expected_output_message in zip(output_messages, expected_output_messages):
        assert output_message.role == expected_output_message.role
        assert output_message.content == expected_output_message.content

def test_two_gptjsons():
    class TestSchema1(BaseModel):
        field1: str
    class TestSchema2(BaseModel):
        field2: str
    
    gptjson1 = GPTJSON[TestSchema1](None)

    # shouldn't allow instantion without a schema
    with pytest.raises(ValueError):
        gptjson2 = GPTJSON(None)
    
    gptjson2 = GPTJSON[TestSchema2](None)

    assert gptjson1.schema_model == TestSchema1
    assert gptjson2.schema_model == TestSchema2



def test_fill_message_template():
    class TestTemplateSchema(BaseModel):
        template_field: str = Field(description="Max length {max_length}")

    gpt = GPTJSON[TestTemplateSchema](None)
    assert gpt.fill_message_template(
        GPTMessage(
            role=GPTMessageRole.USER,
            content="Variable: {max_length}\nMy schema is here: {json_schema}"
        ),
        dict(
            max_length=100,
        )
    ) == GPTMessage(
        role=GPTMessageRole.USER,
        content="Variable: 100\nMy schema is here: {\n\"template_field\": str // Max length 100\n}"
    )

@pytest.mark.asyncio
async def test_extracted_json_is_None():
    gpt = GPTJSON[MySchema](None)

    with patch.object(gpt, 'submit_request', return_value={'choices': [{'message': {'content': 'some content'}}]}), \
         patch.object(gpt, 'extract_json', return_value=(None, FixTransforms(False, False))):
        result, _ = await gpt.run([GPTMessage('system', 'message content')])
        assert result is None

@pytest.mark.asyncio
async def test_no_valid_results_from_remote_request():
    gpt = GPTJSON[MySchema](None)

    with patch.object(gpt, 'submit_request', return_value={'choices': []}):
        result, _ = await gpt.run([GPTMessage('system', 'message content')])
        assert result is None

@pytest.mark.asyncio
async def test_unable_to_find_valid_json_payload():
    gpt = GPTJSON[MySchema](None)

    with patch.object(gpt, 'submit_request', return_value={'choices': [{'message': {'content': 'some content'}}]}), \
         patch.object(gpt, 'extract_json', return_value=(None, FixTransforms(False, False))):
        result, _ = await gpt.run([GPTMessage('system', 'message content')])
        assert result is None

@pytest.mark.asyncio
async def test_unknown_model_to_infer_max_tokens():
    with pytest.raises(ValueError):
        GPTJSON[MySchema](model="UnknownModel", auto_trim=True)



import pytest

from gpt_json.models import ResponseType
from gpt_json.parsers import find_json_response


@pytest.mark.parametrize(
    "input_string,expected,extract_type",
    [
        ('This message is truncated: [{"key1": [123]', '[{"key1": [123]', ResponseType.LIST),
        ('This message is truncated: [{"key1": [123', '[{"key1": [123', ResponseType.LIST),
        ('This message is truncated: [{"key1": "abc"', '[{"key1": "abc"', ResponseType.LIST),
        ('This message is truncated: {"key": "value", "list": [1, 2, 3', '{"key": "value", "list": [1, 2, 3', ResponseType.DICTIONARY),
        ('This message is truncated: {"text": "Test", "numerical": 123, "reason": true, "sub_element": { "name": "Test" }, "items": ["Item 1", "Item 2', '{"text": "Test", "numerical": 123, "reason": true, "sub_element": { "name": "Test" }, "items": ["Item 1", "Item 2', ResponseType.DICTIONARY),
    ]
)
def test_find_json_response(input_string, expected, extract_type):
    assert find_json_response(input_string, extract_type) == expected



def build_stack(json_str):
    stack = []
    fixed_str = ''
    open_quotes = False

    for i, char in enumerate(json_str):
        if not open_quotes:
            # opening a new nested
            if char in "{[":
                stack.append(char)
            # closing a nested
            elif char in "}]":
                stack.pop()
        # opening or closing a string, only it's not escaped
        if char == '"' and i > 0 and json_str[i-1] != "\\":
            open_quotes = not open_quotes

        fixed_str += char

    return (stack, fixed_str, open_quotes)

def is_truncated(json_str):
    """
    Check if the json string is truncated by checking if the number of opening
    brackets is greater than the number of closing brackets.

    """
    stack, _, _ = build_stack(json_str)
    return len(stack) > 0

def fix_truncated_json(json_str):
    """
    Simple json parser that attempts to fix truncated json that might
    be caused by the API response being too long.

    """
    stack, fixed_str, open_quotes = build_stack(json_str)
    is_truncated = len(stack) > 0
    if not is_truncated:
        return json_str, False

    fixed_str = fixed_str.strip()

    if open_quotes:
        fixed_str += '"'

    # Ensure we don't have trailing commas
    fixed_str = fixed_str.strip().rstrip(",")

    # If we still have nested items remaining in our stack,
    # unwind it into the fixed string
    if stack:
        # Unwind the stack by filling it with the closing character
        # of the current nested level
        close_stack = ["]" if char == "[" else "}" for char in stack]
        fixed_str += ''.join(close_stack[::-1])

    return fixed_str, True


def fix_bools(json_str):
    """
    The model will relatively commonly return booleans as capitalized values because of the
    usage of caps in other languages common in the training set (like Python).

    """
    modified = False
    open_quotes = False
    fixed_str = ''

    i = 0
    while i < len(json_str):
        char = json_str[i]

        # Check if the current character is an opening or closing quote
        if char == '"' and i > 0 and json_str[i-1] != "\\":
            open_quotes = not open_quotes

        # If not inside a string, check for "True" or "False" to replace
        if not open_quotes:
            if json_str[i:i+4] == "True":
                fixed_str += "true"
                modified = True
                i += 3  # Skip the remaining characters of "True"
            elif json_str[i:i+5] == "False":
                fixed_str += "false"
                modified = True
                i += 4  # Skip the remaining characters of "False"
            else:
                fixed_str += char
        else:
            fixed_str += char
        i += 1

    return fixed_str, modified



from dataclasses import dataclass
from enum import Enum, unique
import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
    EnumSuper = (StrEnum,)
else:
    EnumSuper = (str, Enum)

@unique
class ResponseType(*EnumSuper):
    DICTIONARY = "DICTIONARY"
    LIST = "LIST"


@unique
class GPTMessageRole(*EnumSuper):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


@unique
class GPTModelVersion(*EnumSuper):
    GPT_3_5 = "gpt-3.5-turbo"
    GPT_4 = "gpt-4-0314"


@dataclass
class FixTransforms:
    """
    How a gpt payload was modified to be valid
    """
    fixed_truncation: bool = False
    fixed_bools: bool = False


@dataclass
class GPTMessage:
    """
    A single message in the chat sequence
    """
    role: GPTMessageRole
    content: str



from re import sub

import pytest

from gpt_json.prompts import generate_schema_prompt
from gpt_json.tests.shared import MySchema


def strip_whitespace(input_string: str):
    return sub(r"\s+", "", input_string)

@pytest.mark.parametrize(
    "schema_definition,expected",
    [
        (
            MySchema,
            """
            {{
                "text": str,
                "items": str[],
                "numerical": int | float,
                "sub_element": {{
                    "name": str
                }},
                "reason": bool // Explanation
            }}
            """
        ),
        (
            list[MySchema],
            """
            [
            {{
            "text": str,
            "items": str[],
            "numerical": int | float,
            "sub_element": {{
                "name": str
            }},
            "reason": bool // Explanation
            }}, // Repeat for as many objects as are relevant
            ]
            """
        )
    ]
)
def test_generate_schema_prompt(schema_definition, expected: str):
    assert strip_whitespace(generate_schema_prompt(schema_definition)) == strip_whitespace(expected)
